import random

# Selection algorithms for choosing clusters
# Each selection algorithm should return a list


def random_selection(choices):
    """
    Select randomly
    """
    return [random.choice(choices)]
